// PxM.cpp
// (c) 2001-2004 by Jeremy Stanley
// This is free sofwtare licensed under the GNU GPL; see license.txt for details.

#include "PxM.h"

#ifdef USE_GRAPHICS_H
#include "graphics.h"
#endif

#ifdef USE_JPEG_LIB
extern "C" {
#include <jpeglib.h>
};
#endif

#include <assert.h>

#ifdef _MSC_VER
#define strcasecmp _stricmp
#undef create_compress
#undef create_decompress
#endif

const int TOKEN = 128;

// PxM image functions

// default constructor
PxM::PxM()
{
    pixels = NULL;
    width = 0;
    height = 0;
    iscolor = 0;
    bytes = 0;
    bytes_per_pixel = 0;
    pitch = 0;
}

// copy constructor
PxM::PxM(const PxM &other)
{
	pixels = NULL;
    *this = other;
}

// new-image constructor
PxM::PxM(int w, int h, bool color /*= true*/)
{
	pixels = NULL;
	init(width, height, color);
}

void PxM::init(int w, int h, bool color /* = true */ )
{
	reset();
	
	width = w;
	height = h;
	if (color)
	{
		iscolor = 1;
		bytes_per_pixel = 3;
	}
	else
	{
		iscolor = 0;
		bytes_per_pixel = 1;
	}
	pitch = width * bytes_per_pixel;
	bytes = pitch * height;
	pixels = new unsigned char[bytes];
	memset(pixels, 0, bytes);
}

void PxM::reset()
{
	width = 0;
	height = 0;
	bytes = 0;
	pitch = 0;
	if ( pixels )
	{
		delete [] pixels;
		pixels = NULL;
	}
}

// destructor
PxM::~PxM()
{
    reset();
}

// gets a whitespace-delimited PPM/PGM token
char *PxM::getToken(char *buf, int size, FILE *stream)
{
    char c, *p = buf;
    int sz = 0;

    // eat initial whitespace
    do {
        c = fgetc(stream);
    } while(c != EOF && isspace(c));

    if(c == EOF) return NULL;

    // eat comments (if applicable)
    while(c == '#')
    {
        // eat the comment
        do {
            c = fgetc(stream);
        } while(c != '\n' && c != EOF);
        if(c == EOF) return NULL;

        // eat trailing whitespace
        do {
            c = fgetc(stream);
        } while(c != EOF && isspace(c));
    }

    if(c == EOF) return NULL;

    // retrieve token
    do
    {
        if(++sz < size)    // save room for terminator
            *p++ = c;
        c = fgetc(stream);
    } while(c != EOF && !isspace(c));

    // null-terminate token
    *p = '\0';

    return buf;
}

// Loads any type of supported file, detected by extension
int PxM::load(const char *filename)
{
	const char *ext = strrchr(filename, '.');
	if (!ext) return 0;

	if (0 == strcasecmp(ext, ".ppm") || 0 == strcasecmp(ext, ".pgm"))
		return load_PxM(filename);

#ifdef USE_JPEG_LIB
	if (0 == strcasecmp(ext, ".jpg") || 0 == strcasecmp(ext, ".jpeg"))
		return load_jpeg(filename);
#endif

	return 0;
}

// Loads a PPM or PGM file into memory
int PxM::load_PxM(const char *filename)
{
	FILE *f = fopen(filename, "rb");
	if(!f) {
		fprintf(stderr, "PxM::load_PxM(): Couldn't open %s\n", filename);
		return 0;
	}
	int res = load_PxM(f);
	fclose(f);
	return res;
}

int PxM::load_PxM(FILE *f)
{
    reset();

    static char tokenbuf[TOKEN];

	int filetype;
	getToken(tokenbuf, TOKEN, f);
    if(0==strcmp(tokenbuf, "P2"))  // text PGM
    {
        filetype = 2;
        iscolor = 0;
        bytes_per_pixel = 1;
    }
    else if(0==strcmp(tokenbuf, "P5")) // binary PGM
    {
        filetype = 5;
        iscolor = 0;
        bytes_per_pixel = 1;
    }
    else if(0==strcmp(tokenbuf, "P3")) // text PPM
    {
        filetype = 3;
        iscolor = 1;
        bytes_per_pixel = 3;
    }
    else if(0==strcmp(tokenbuf, "P6")) // binary PPM
    {
        filetype = 6;
        iscolor = 1;
        bytes_per_pixel = 3;
    }
    else
    {
        fprintf(stderr, "PxM::load_PxM(): Type \"%s\" is not supported", tokenbuf);
        return 0;
    }

    getToken(tokenbuf, TOKEN, f);
    width = atoi(tokenbuf);

    getToken(tokenbuf, TOKEN, f);
    height = atoi(tokenbuf);

    getToken(tokenbuf, TOKEN, f);
    int bcc = atoi(tokenbuf);

    //printf("%s: Type=%d  width=%d  height=%d  max color component=%d\n",
    //       filename, filetype, width, height, bcc);
    if(bcc != 255) {
        fprintf(stderr, "PxM::load_PxM(): This implementation only supports 8-bit color component depths!\n");
        return 0;
    }
    fflush(stdout);

    pitch = width * bytes_per_pixel;
    bytes = pitch * height;

    pixels = new unsigned char[bytes];
    unsigned char *p = pixels;

    for(int i = 0; i < bytes; i++)
    {
        if(filetype == 5 || filetype == 6)
        {
            // read binary
            *p++ = fgetc(f);
        }
        else
        {
            // read text
            getToken(tokenbuf, TOKEN, f);
            *p++ = atoi(tokenbuf);
        }
    }

    return 1;
}

// Saves the image as PGM or PPM, whichever is appropriate, using the binary (raw bits) format
int PxM::save_PxM(const char *filename)
{
    if(!pixels) {
        fprintf(stderr, "PxM::save_PxM():  Image not initialized\n");
        return 0;
    }

    FILE *f = fopen(filename, "wb");
    if(!f) {
        fprintf(stderr, "PxM::save_PxM():  Couldn't open %s\n", filename);
        return 0;
    }
    if(iscolor)
        fprintf(f, "P6\n");
    else
        fprintf(f, "P5\n");
    fprintf(f, "%d %d\n%d\n", width, height, 255);

    for(int i = 0; i < bytes; i++)
        fputc(pixels[i], f);

    fclose(f);

    return 1;
}

// Converts the image to greyscale
int PxM::convertbw()
{
    reset();

    if(!iscolor || !pixels)
        return 0;

    unsigned char *tmp = new unsigned char[width * height];
    if(!tmp) return 0;

    int i, j = 0;
    for(i = 0; i < bytes; i += 3)
        tmp[j++] = (int)(0.299 * pixels[i] + 0.587 * pixels[i + 1] + 0.114 * pixels[i + 2]);

    pixels = tmp;
    bytes_per_pixel = 1;
    pitch = width;
    bytes = width * height;
    iscolor = 0;

    return 1;
}

// Assigns an image to another image
PxM &PxM::operator=(const PxM &other)
{
    reset();
	width = other.width;
    height = other.height;
	pitch = other.pitch;
    iscolor = other.iscolor;
    bytes = other.bytes;
	bytes_per_pixel = other.bytes_per_pixel;
    pixels = new unsigned char[bytes];
    memcpy(pixels, other.pixels, bytes);
    return *this;
}

// Subtracts another image from this
int PxM::subtract(PxM &other)
{
    if(!pixels || !other.pixels || bytes != other.bytes)
        return 0;

    for(int i=0; i < bytes; i++)
        pixels[i] -= other.pixels[i];

    return 1;
}

// Negates an image
int PxM::negate()
{
    if(!pixels)
        return 0;

    for(int i=0; i < bytes; i++)
        pixels[i] = ~pixels[i];

    return 1;
}

void PxM::setpixel(int u, int v, const int col[3])
{
	assert( !(u < 0 || u >= width || v < 0 || v >= height) );
	if ( u < 0 || u >= width || v < 0 || v >= height )
		return;

	unsigned char *pix = pixels + v * pitch + u * bytes_per_pixel;
	assert( (pix >= pixels) && (pix + bytes_per_pixel <= pixels + bytes) );
	for(int i = 0; i < bytes_per_pixel; i++)
	{
		pix[i] = col[i];
	}
}

void PxM::pixel(int x, int y, int col[3])
{
	if(!pixels || x < 0 || x >= width || y < 0 || y >= height)
	{
		col[0] = 255;
		col[1] = col[2] = 0;
		return;
	}

	unsigned char *pix = pixels + y * pitch + x * bytes_per_pixel;
	
	if(iscolor) {
		col[0] = pix[0];
		col[1] = pix[1];
		col[2] = pix[2];
	}
	else {
		col[0] = col[1] = col[2] = pix[0];
	}
}

void PxM::nearestpix(double u, double v, int col[3])
{
    unsigned int x = (unsigned int)u, y = (unsigned int)v;
    unsigned char *p = pixels + pitch * y + bytes_per_pixel * x;
    if(iscolor) {
        col[0] = p[0]; col[1] = p[1]; col[2] = p[2];
    }
    else {
        col[0] = col[1] = col[2] = p[0];
    }
}

void PxM::bilinearpix(double u, double v, int col[3])
{
    unsigned int x = (unsigned int)u, y = (unsigned int)v;
    int i;
    unsigned char c00[3], c01[3], c10[3], c11[3];
    double dx = u - x;
    double dy = v - y;
    double fx0, fx1;

    // Error:  Just return a red pixel (not likely to occur naturally in my grayscale test images :)
    if(x >= (unsigned int)width || y >= (unsigned int)height)
    {
        col[0] = 255;
        col[1] = 0;
        col[2] = 0;
        return;
    }

    // ensure we have pixels both down and to the right to work with
    if(x == (unsigned int)width-1) { x--; dx = 1.0f; }
    if(y == (unsigned int)height-1) { y--; dy = 1.0f; }

    // Now... we have the pixels at (x, y) .. (x+1, y+1) to interpolate about
    // dx = weight toward (x+1); dy = weight toward (y+1)

    // Find pointer to upper-left pixel
    unsigned char *p = pixels + pitch * y + bytes_per_pixel * x;

    // Get colors of 4 neighbors
    memcpy(c00, p, bytes_per_pixel);
    memcpy(c10, p + bytes_per_pixel, bytes_per_pixel);
    memcpy(c01, p + pitch, bytes_per_pixel);
    memcpy(c11, p + bytes_per_pixel + pitch, bytes_per_pixel);

    // Perform bilinear interpolation
    for(i = 0; i < bytes_per_pixel; i++)
    {
        fx0 = c00[i] + dx * (c10[i] - c00[i]);
        fx1 = c01[i] + dx * (c11[i] - c01[i]);
        col[i] = (int)(fx0 + dy * (fx1 - fx0));
    }

    // Fill unused array entries for greyscale pictures
    if(!iscolor)
    {
        col[1] = col[0];
        col[2] = col[0];
    }
}

int PxM::scale(int new_width, int new_height)
{
	if (new_width <= 0 || new_height <= 0) return 0;

	unsigned char *new_pixels = new unsigned char [new_width * new_height * bytes_per_pixel];
	if (!new_pixels) return 0;

	int x, y, i, c[3];
	double u = 0.0, v = 0.0;
	double du = (double)width / new_width;
	double dv = (double)height / new_height;

	unsigned char *p = new_pixels;

	for(y = 0; y < new_height; y++)
	{
		u = 0.0;

		for(x = 0; x < new_width; x++)
		{
			bilinearpix(u, v, c);

			for(i = 0; i < bytes_per_pixel; i++)
			{
				p[i] = c[i];
			}
			p += bytes_per_pixel;

			u += du;
		}

		v += dv;
	}

	reset();
	pixels = new_pixels;
	width = new_width;
	height = new_height;
	pitch = new_width * bytes_per_pixel;
	bytes = pitch * height;

	return 1;
}

#ifdef USE_JPEG_LIB

// Loads a JPEG file into memory
int PxM::load_jpeg(const char *filename)
{
    reset();

    FILE *f = fopen(filename, "rb");
    if(!f) {
        fprintf(stderr, "PxM::load_jpeg(): Couldn't open %s\n", filename);
        return 0;
	}

	jpeg_decompress_struct cinfo;
	jpeg_error_mgr jerr;
	cinfo.err = jpeg_std_error(&jerr);

	jpeg_create_decompress(&cinfo);
	
	jpeg_stdio_src(&cinfo, f);
	jpeg_read_header(&cinfo, TRUE);
	jpeg_start_decompress(&cinfo);

	bytes_per_pixel = cinfo.out_color_components;
	iscolor = (bytes_per_pixel > 1);
	width = cinfo.output_width;
	height = cinfo.output_height;

    pitch = width * bytes_per_pixel;
    bytes = pitch * height;

    pixels = new unsigned char[bytes];
    unsigned char *p = pixels;

	while(cinfo.output_scanline < cinfo.output_height)
	{
		jpeg_read_scanlines(&cinfo, &p, 1);
		p += pitch;
	}

	jpeg_finish_decompress(&cinfo);
	jpeg_destroy_decompress(&cinfo);

    fclose(f);

    return 1;
}

// Saves the image as JPEG
int PxM::save_jpeg(const char *filename, int quality)
{
    if(!pixels) {
        fprintf(stderr, "PxM::save_jpeg():  Image not initialized\n");
        return 0;
    }

    FILE *f = fopen(filename, "wb");
    if(!f) {
        fprintf(stderr, "PxM::save_jpeg():  Couldn't open %s\n", filename);
        return 0;
    }

	jpeg_compress_struct cinfo;
	jpeg_error_mgr jerr;
	cinfo.err = jpeg_std_error(&jerr);

	jpeg_create_compress(&cinfo);
	
	jpeg_stdio_dest(&cinfo, f);

	cinfo.in_color_space = iscolor ? JCS_RGB : JCS_GRAYSCALE;
	jpeg_set_defaults(&cinfo);

	cinfo.image_width = width;
	cinfo.image_height = height;
	cinfo.input_components = bytes_per_pixel;

	jpeg_set_quality(&cinfo, quality, FALSE);

	jpeg_start_compress(&cinfo, TRUE);

	unsigned char *row_ptr = pixels;

	while(cinfo.next_scanline < cinfo.image_height)
	{
		jpeg_write_scanlines(&cinfo, &row_ptr, 1);
		row_ptr += pitch;
	}

	jpeg_finish_compress(&cinfo);
	jpeg_destroy_compress(&cinfo);

	fclose(f);

    return 1;
}

#endif // USE_JPEG_LIB

#ifdef USE_GRAPHICS_H

// Captures the w * h region with upper-left corner (x0, y0)
int PxM::get(int x0, int y0, int w, int h, int c)
{
    reset();

    width = w;
    height = h;
    iscolor = c;
    bytes = 0;

    if(iscolor)
        bytes_per_pixel = 3;
    else
        bytes_per_pixel = 1;

    pitch = width * bytes_per_pixel;
    bytes = pitch * height;

    pixels = new unsigned char[bytes];

    unsigned char *pDisp = (unsigned char *)pSDL->pixels + y0 * pSDL->pitch + x0 * Bpp;
    unsigned char *pImg = pixels;
    int x, y, r, g, b, dd = pSDL->pitch - (Bpp * w);

    for(y = 0; y < h; y++)
    {
        for(x = 0; x < w; x++)
        {
            r = pDisp[R_off];
            g = pDisp[G_off];
            b = pDisp[B_off];

            if(iscolor) {
                pImg[0] = r;
                pImg[1] = g;
                pImg[2] = b;
            }
            else
            {
                pImg[0] = (int)(0.299 * r + 0.587 * g + 0.114 * b);
            }

            pImg += bytes_per_pixel;
            pDisp += Bpp;
        }

        pDisp += dd;
    }

    return 1;
}


// Draws the image at (x0, y0) at its natural height
// Clips if necessary
void PxM::draw(int x0, int y0)
{
    int tR, tG, tB;
    if(iscolor) {
        tR = 0; tG = 1; tB = 2;
    }
    else {
        tR = 0; tG = 0; tB = 0;
    }

    int w = width;
    int h = height;
    unsigned char *pTex;

    // clip the left edge if needed
    if(x0 < 0) {
        w += x0;
        pTex = pixels - x0 * bytes_per_pixel;
        x0 = 0;
    }
    else pTex = pixels;

    // clip the top edge if needed
    if(y0 < 0) {
        h += y0;
        pTex -= y0 * pitch;
        y0 = 0;
    }

    Uint8 *pPix = (Uint8 *)pSDL->pixels + pSDL->pitch * y0 + Bpp * x0;
    Uint8 *pTexRow;
    Uint8 *pPixRow;

    int j;

    // compute lower right screen coordinates
    int x1 = x0 + w - 1;
    int y1 = y0 + h - 1;

    // clip bottom and right edges if needed
    if(x1 >= W_WIDTH) x1 = W_WIDTH - 1;
    if(y1 >= W_HEIGHT) y1 = W_HEIGHT - 1;

    for(; y0 <= y1; y0++)
    {
        pPixRow = pPix;
        pTexRow = pTex;

        // Write a row
        for(j = x0; j <= x1; j++)
        {
            // Write a pixel...
            pPixRow[R_off] = pTexRow[tR];
            pPixRow[G_off] = pTexRow[tG];
            pPixRow[B_off] = pTexRow[tB];

            // ...and advance the screen pointer to the next column
            pPixRow += Bpp;

            // Advance the texture pointer to the next column
            pTexRow += bytes_per_pixel;
        }

        // Advance the screen pointer to the next row
        pPix += pSDL->pitch;

        // Advance the texture pointer to the next row
        pTex += pitch;
    }
}

// Centers the image on (x0, y0)
void PxM::plot_center(int x0, int y0)
{
	draw(x0 - width/2, y0 - height/2);
}

// Draws the image at (x0, y0) with width w, height h
// Clips if necessary
void PxM::draw(int x0, int y0, int w, int h)
{
    int tR, tG, tB;
    if(iscolor) {
        tR = 0; tG = 1; tB = 2;
    }
    else {
        tR = 0; tG = 0; tB = 0;
    }

    int tw = width;
    int th = height;
    unsigned char *pTex;

    // clip the left edge if needed
    if(x0 < 0) {
        int dt = (x0 * width) / w;
        w += x0;
        tw += dt;
        pTex = pixels - dt * bytes_per_pixel;
        x0 = 0;
    }
    else pTex = pixels;

    // clip the top edge if needed
    if(y0 < 0) {
        int dt = (y0 * height) / h;
        h += y0;
        th += dt;
        pTex -= dt * pitch;
        y0 = 0;
    }

    Uint8 *pPix = (Uint8 *)pSDL->pixels + pSDL->pitch * y0 + Bpp * x0;
    Uint8 *pTexRow;
    Uint8 *pPixRow;

    // Now we have a (tw x th) pixmap to be mapped to a (w x h) screen window
    int j;
    int n0 = th-1, d0 = h-1, i0 = d0;
    int num, den, inc;

    // compute lower right screen coordinates
    int x1 = x0 + w - 1;
    int y1 = y0 + h - 1;

    // clip bottom and right edges if needed
    if(x1 >= W_WIDTH) x1 = W_WIDTH - 1;
    if(y1 >= W_HEIGHT) y1 = W_HEIGHT - 1;

    for(; y0 <= y1; y0++)
    {
        pPixRow = pPix;
        pTexRow = pTex;

        // Write a row
        num = tw-1; den = w-1; inc = den;
        for(j = x0; j <= x1; j++)
        {
            // Write a pixel...
            pPixRow[R_off] = pTexRow[tR];
            pPixRow[G_off] = pTexRow[tG];
            pPixRow[B_off] = pTexRow[tB];

            // ...and advance the screen pointer
            pPixRow += Bpp;

            // Handle single-pixel width correctly
            if(den==0) break;

            // Advance the texture pointer if needed
            inc += num;
            while(inc > den)
            {
                pTexRow += bytes_per_pixel;
                inc -= den;
            }
        }

        // Handle single-pixel height correctly
        if(d0==0) break;

        // Advance the screen pointer
        pPix += pSDL->pitch;

        // Advance the texture pointer if needed
        i0 += n0;
        while(i0 > d0)
        {
            pTex += pitch;
            i0 -= d0;
        }
    }
}

// plots the image centered at (x, y), rotated about its center by angle degrees
void PxM::plot_rotate(int x, int y, double angle)
{
    // set default axis of rotation
    // (rotate about center of image)
    int p = width / 2, q = height / 2;

    // compute vertices of destination quadrilateral
    int verts[4][2];
    int rverts[4][2];

    verts[0][0] = verts[3][0] = -p; // left
    verts[0][1] = verts[1][1] = -q; // top
    verts[3][1] = verts[2][1] = height - q - 1; // bottom
    verts[1][0] = verts[2][0] = width - p - 1; // right

    // scale and rotate destination quadrilateral
    // x' = (x - xc) cos t - (y - yc) sin t + xc
    // y' = (x - xc) sin t + (y - yc) cos t + yc
    double t = angle * (3.14159265359 / 180.0);
    double c = cos(t), s = sin(t);

    int i;
    for(i = 0; i < 4; i++)
    {
        rverts[i][0] = x + (int)(verts[i][0] * c - verts[i][1] * s);
        rverts[i][1] = y + (int)(verts[i][0] * s + verts[i][1] * c);
    }

    // find the minimum and maximum extents of the resultant image
    int min_x = rverts[0][0], max_x = rverts[0][0];
    int min_y = rverts[0][1], max_y = rverts[0][1];
    for(i = 1; i < 4; i++)
    {
        if(rverts[i][0] > max_x) max_x = rverts[i][0];
		if(rverts[i][0] < min_x) min_x = rverts[i][0];
		if(rverts[i][1] > max_y) max_y = rverts[i][1];
		if(rverts[i][1] < min_y) min_y = rverts[i][1];
    }

    // plot the rotated image
    // ----------------------
    int X, Y, u = (min_x + max_x) / 2, v = (min_y + max_y) / 2;
    int col[3];
    double px, py;

    // Negate the angle since this is a reverse mapping
    t = -t;
    c = cos(t), s = sin(t);

    // Iterate through each pixel in the output
    for(Y = min_y; Y <= max_y; Y++)
    {
        for(X = min_x; X <= max_x; X++)
		{
	    	// map into the image
	    	px = ((X - u) * c - (Y - v) * s) + p;
	    	py = ((X - u) * s + (Y - v) * c) + q;

	    	// check the bounds
	    	if(px < 0.0f || py < 0.0f || px >= (double)width || py >= (double)height)
		    	continue;

	    	// get the color
	    	bilinearpix(px, py, col);

	    	// plot the pixel
            plotpixel(X, Y, col[0], col[1], col[2]);
        }
    }
}

#endif // USE_GRAPHICS_H
