// ps-sdl_gl_app.cpp
// (c) 2004 by Jeremy Stanley
// This is free software licensed under the GNU GPL.  See license.txt for details.

#include "ps-sdl_gl_app.h"
#include "ps-timer.h"

const int 	TIMESTAMP_COUNT = 4;		// frame time is averaged over this many frames
const int	MY_SDL_INVALIDATE = SDL_USEREVENT + 1;
const int	INITIAL_FRAMERATE  = 50;	// initial FPS; a rough guess.  it will be correct soon enough.

psSdlGlApp *theApp;

psSdlGlApp::psSdlGlApp()
{
	m_Frametime = 1.0 / INITIAL_FRAMERATE;	// this is adjusted by calibrate_speed(), which is called once per frame
	m_timestamps = 0;
	m_busy = 0;
	m_quit_req = false;
	m_need_redraw = false;

	theApp = this;
}

psSdlGlApp::~psSdlGlApp()
{
	SDL_Quit();
}

bool psSdlGlApp::InitViewport(int width, int height, int depth, int mode, const char *caption)
{
	// setup window parameters
	m_Width = width;
	m_Height = height;
	m_mode = mode;
	m_caption = caption;

	/* todo: create window ... */

    /* First, initialize SDL's video subsystem. */
    if( SDL_Init( SDL_INIT_VIDEO | SDL_INIT_TIMER | SDL_INIT_AUDIO ) < 0 ) {
        /* Failed, exit. */
        fprintf( stderr, "Video initialization failed: %s\n",
             SDL_GetError( ) );
        return false;
    }

    /* Let's get some video information. */
    const SDL_VideoInfo* info = SDL_GetVideoInfo( );

    if( !info ) {
        /* This should probably never happen. */
        fprintf( stderr, "Video query failed: %s\n",
             SDL_GetError( ) );
        return false;
    }

	/* Set bit depth to desktop resolution if not in "game mode" */
	if (mode < 2)
	{
		depth = info->vfmt->BitsPerPixel;
	}

    /* Set GL attributes */
    SDL_GL_SetAttribute( SDL_GL_DOUBLEBUFFER, 1 );
	if (depth < 32)
	{
		SDL_GL_SetAttribute( SDL_GL_RED_SIZE, 5 );
		SDL_GL_SetAttribute( SDL_GL_GREEN_SIZE, 5 );
		SDL_GL_SetAttribute( SDL_GL_BLUE_SIZE, 5 );
		SDL_GL_SetAttribute( SDL_GL_DEPTH_SIZE, 16 );
	}
	else
	{
		SDL_GL_SetAttribute( SDL_GL_RED_SIZE, 8 );
		SDL_GL_SetAttribute( SDL_GL_GREEN_SIZE, 8 );
		SDL_GL_SetAttribute( SDL_GL_BLUE_SIZE, 8 );
		SDL_GL_SetAttribute( SDL_GL_DEPTH_SIZE, 24 );
	}

	/* Set the video mode */
    int flags = SDL_OPENGL;
	if (mode == 0)
		flags |= SDL_RESIZABLE;
	else if (mode == 1)
	{
		flags |= SDL_NOFRAME;
	}
	else
		flags |= SDL_FULLSCREEN;

	if( ! SDL_SetVideoMode( width, height, depth, flags) ) {
        /*
         * This could happen for a variety of reasons,
         * including DISPLAY not being set, the specified
         * resolution not being available, etc.
         */
        fprintf( stderr, "Video mode set failed: %s\n",
             SDL_GetError( ) );
        return false;
    }

	/* Set the window caption */
	SDL_WM_SetCaption(m_caption.c_str(), m_caption.c_str());

	return true;
}

int psSdlGlApp::ProcessEvents()
{
	SDL_Event event;

	if (m_busy > 0)
	{
		// process any events that may be waiting
		while (1 == SDL_PollEvent(&event))
		{
			// dispatch the event
			DispatchEvent(event);
		}

		// update game state based on elapsed frame time
		UpdateBusyState(m_Frametime);

		// refresh the display
		do_display();
	}
	else
	{
		// wait for an event
		if (1 == SDL_WaitEvent(&event))
		{
			// dispatch the event
			DispatchEvent(event);
		}
	}

	return 0;
}

void psSdlGlApp::Run()
{
	// initialize the timer
	if (!psInitializeTimer())
	{
		fprintf(stderr, "psSdlGlApp::Run(): Couldn't initialize timer!\n");
		return;
	}

	SDL_EnableKeyRepeat(SDL_DEFAULT_REPEAT_DELAY, SDL_DEFAULT_REPEAT_INTERVAL);
	SDL_EnableUNICODE(true);

	while(!m_quit_req)
	{
		ProcessEvents();
	}
}

void psSdlGlApp::Invalidate()
{
	// SDL_UpdateRect doesn't appear to work in OpenGL modes; it crashes and burns...
	if ( !m_need_redraw )
	{
		SDL_Event ev;
		memset(&ev, 0, sizeof(SDL_Event));
		ev.type = MY_SDL_INVALIDATE;
		SDL_PushEvent(&ev);
		
		m_need_redraw = true;
	}
}

void psSdlGlApp::DispatchEvent(SDL_Event &event)
{
	switch(event.type)
	{
	case SDL_KEYDOWN:
		KeyDown(event.key.keysym.sym, event.key.keysym.mod, event.key.keysym.unicode);
		break;
	case SDL_KEYUP:
		KeyUp(event.key.keysym.sym, event.key.keysym.mod);
		break;
	case SDL_VIDEORESIZE:
#ifdef WIN32
		{
			// Holy cow, this is evil.
			// But SDL_SetVideoMode() will clobber my OpenGL state on Win32.
			SDL_Surface *surface = SDL_GetVideoSurface();
			surface->w = event.resize.w;
			surface->h = event.resize.h;
			surface->pitch = surface->w * surface->format->BytesPerPixel;

			// Note: after actually looking at the surface, and seeing that pixels is NULL
			// and hwdata is NULL, I don't think this is harmful.
		}
#else
		SDL_SetVideoMode(event.resize.w, event.resize.h, 0, SDL_OPENGL | SDL_RESIZABLE);
#endif
		glViewport(0, 0, (GLsizei) event.resize.w, (GLsizei) event.resize.h);
		m_Width = event.resize.w;
		m_Height = event.resize.h;
		Reshape(m_Width, m_Height);
		Invalidate();
		break;
	case MY_SDL_INVALIDATE:
	case SDL_VIDEOEXPOSE:
		if (m_busy <= 0)
			do_display();
		break;
	case SDL_MOUSEBUTTONDOWN:
		MouseDown(event.button.button, event.button.x, event.button.y);
		break;
	case SDL_MOUSEBUTTONUP:
		MouseUp(event.button.button, event.button.x, event.button.y);
		break;
	case SDL_MOUSEMOTION:
		MouseMove(event.motion.state, event.motion.x, event.motion.y, event.motion.xrel, event.motion.yrel);
		break;
	case SDL_QUIT:
		m_quit_req = OnQuit();
		break;
	case SDL_USEREVENT:
		OnUserEvent(event.user.code, event.user.data1, event.user.data2);
		break;
	}
}

void psSdlGlApp::calibrate_speed()
{
	// speed calibration code
	// sets m_Frametime to seconds per frame, averaged over the last TIMESTAMP_COUNT frames

	static ps_tick_t timestamp[TIMESTAMP_COUNT];

	if (m_timestamps < TIMESTAMP_COUNT)
	{
		timestamp[m_timestamps++] = psGetTickCount();
	}
	else
	{
		memmove(timestamp, timestamp + 1, (TIMESTAMP_COUNT - 1) * sizeof(ps_tick_t));
		timestamp[TIMESTAMP_COUNT - 1] = psGetTickCount();
	}

	if (m_timestamps > 1)
	{
		m_Frametime = psGetSecondsElapsed(timestamp[m_timestamps - 1] - timestamp[0]) / (m_timestamps - 1);
	}
}

// While the application is in a "busy" state, your application gets continuous
// UpdateBusyState() and Draw() calls.  Otherwise, UpdateBusyState() is not issued and
// Draw() is issued only when the viewport is un-obscured or resized.

void psSdlGlApp::BeginBusyState()
{
	if (m_busy == 0)
	{
		m_timestamps = 0;
	}
	m_busy++;
}

void psSdlGlApp::EndBusyState()
{
	m_busy--;
}

void psSdlGlApp::ResetBusyState()
{
	m_busy = 0;
}

void psSdlGlApp::PostQuit()
{
	SDL_Event ev;
	ev.type = SDL_QUIT;
	SDL_PushEvent(&ev);
}

void psSdlGlApp::do_display(void)
{
	// draw game board
	Draw();

	// if in busy state, calibrate speed
	if (m_busy > 0)
	{
		calibrate_speed();
	}

	// do it!
	SDL_GL_SwapBuffers();
	
	m_need_redraw = false;
}

