// timer.cpp
// (c) 2004 by Jeremy Stanley
// This is free software licensed under the GNU GPL.  See license.txt for details.
//////////////////////////////////////////////////////////////////////

#include "ps-timer.h"
#include <stdio.h>

#if defined( WIN32 )

/************************************
 * win32 implementation             *
 ************************************/

// JS 9/22/04: stopped using RDTSC, as it's incompatible with power saving speed throttling...

//#define USE_RDTSC	// Use the Pentium timestamp counter.  Less overhead than the performance counter,
					// but only available on Pentium (I don't check), requires us to measure the
					// CPU speed initially, and doesn't work if the CPU speed can change during
					// execution (as when a laptop goes on battery power, for example...)

#define WIN32_LEAN_AND_MEAN		// Exclude rarely-used stuff from Windows headers

#include <windows.h>

#ifdef USE_RDTSC

#include <mmsystem.h>

static const CALIBRATE_INTERVAL_MS = 200;
static volatile ps_tick_t startCount, endCount;
static double Hz;

void CALLBACK TimeProc(UINT uID, UINT uMsg, DWORD d1, DWORD d2, DWORD d3)
{
	if(startCount==0)
	{
		startCount = psGetTickCount();
		return;
	}

	endCount = psGetTickCount();
	timeKillEvent(uID);
	return;
}

bool psInitializeTimer()
{
	startCount = endCount = 0;

	UINT ev = timeSetEvent(CALIBRATE_INTERVAL_MS, 0, TimeProc, 0, TIME_PERIODIC | TIME_CALLBACK_FUNCTION);
	if (!ev)
	{
		fprintf(stderr, "Could not create timer event\n");
		return false;
	}

	while(endCount == 0);

	ps_tick_t ticks = endCount - startCount;

	Hz = (double)ticks * (1000.0 / (double)CALIBRATE_INTERVAL_MS);
	fprintf(stderr, "Computed CPU speed: %f MHz\n", Hz / 1.0e6);

	return true;
}

ps_tick_t psGetTickCount()
{
	ps_tick_t S;

	__asm rdtsc;
	__asm mov dword ptr[S], eax;
	__asm mov dword ptr[S + 4], edx;

	return S;
}

double psGetSecondsElapsed(ps_tick_t ticks)
{
	return (double)ticks / Hz;
}

#else

static double Hz;

bool psInitializeTimer()
{
	LARGE_INTEGER freq;

	if (!QueryPerformanceFrequency(&freq))
	{
		fprintf(stderr, "System does not support a high-resolution performance counter.\n");
		return false;
	}

	Hz = (double)freq.QuadPart;

	return true;
}

ps_tick_t psGetTickCount()
{
	LARGE_INTEGER res;
	QueryPerformanceCounter(&res);
	return (ps_tick_t)(res.QuadPart);
}

double psGetSecondsElapsed(ps_tick_t ticks)
{
	return (double)ticks / Hz;
}

#endif

#else

/************************************
 * Unix-y implementation            *
 ************************************/
 
#include <sys/time.h>

bool psInitializeTimer()
{
	return true;
}

ps_tick_t psGetTickCount()
{
	timeval tv;
	gettimeofday(&tv, NULL);
	return (ps_tick_t)tv.tv_sec * 1000000 + tv.tv_usec;
}

double psGetSecondsElapsed(ps_tick_t ticks)
{
	return (double)ticks / 1.0e6;
}

#endif
